import theano
import theano.gradient
import theano.printing
import theano.gof
from theano import gof
from theano.sandbox.cuda.basic_ops import (as_cuda_ndarray_variable, gpu_contiguous)
from .Util import raw_variable
from theano.gof.opt import OpSub
from theano.compile import optdb
import theano.tensor as T
from .Util import get_c_support_code_common, get_c_support_code_mdlstm




class MultiDirectionalTwoDLSTMOpGrad(theano.sandbox.cuda.GpuOp):
  __props__ = ("inplace",)

  def __init__(self, inplace):
    super(MultiDirectionalTwoDLSTMOpGrad, self).__init__()
    self.inplace = inplace
    if inplace:
      #inputs: X, 4xW, 4xV_h, 4xV_v, 4xb, 4xDy, 4xY, 4xH

      #note: outputs [17,18,19,20] are the DYs, but we don't use them here for inplace,
      #as they will usually be aligned with each other (atleast when the sum of the outputs is used)

      #all outputs operate inplace on inputs [26,27,28,29] (which are the Hs)
      #but when the input is marked multiple times, we get an error
      #so we do this workaround
      #anyway theano knows that these inputs will be destroyed, so it should be OK
      self.destroy_map = {4: [26], 5: [27], 6: [28], 7: [29]}

  def make_node(self, X, W1, W2, W3, W4, V_h1, V_h2, V_h3, V_h4, V_v1, V_v2, V_v3, V_v4,
                b1, b2, b3, b4, sizes, DY1, DY2, DY3, DY4, Y1, Y2, Y3, Y4, H1, H2, H3, H4):

    var_names = ["X", "W1", "W2", "W3", "W4", "V_h1", "V_h2", "V_h3", "V_h4",
                 "V_v1", "V_v2", "V_v3", "V_v4", "b1", "b2", "b3", "b4",
                 "DY1", "DY2", "DY3", "DY4", "Y1", "Y2", "Y3", "Y4",
                 "H1", "H2", "H3", "H4"]
    lcl = locals()
    for var_name in var_names:
      lcl[var_name] = gpu_contiguous(as_cuda_ndarray_variable(lcl[var_name]))
      assert lcl[var_name].dtype == "float32"
    #note: sizes lives on the CPU!
    sizes = T.as_tensor_variable(sizes)
    assert sizes.dtype == "float32"
    expected_ndims = [4] + ([2] * 12) + ([1] * 4) + ([4] * 12)
    assert len(var_names) == len(expected_ndims), (len(var_names), len(expected_ndims))
    for var_name, expected_ndim in zip(var_names, expected_ndims):
      assert lcl[var_name].ndim == expected_ndim, \
          (var_name, lcl[var_name].name, lcl[var_name].ndim, expected_ndim)
    assert sizes.ndim == 2
    all_vars_no_sizes = [lcl[var_name] for var_name in var_names]
    all_vars = all_vars_no_sizes[:17] + [sizes] + all_vars_no_sizes[17:]
    inputs_vars = all_vars[:17]

    return theano.Apply(self, all_vars, [v.type() for v in inputs_vars])

  def c_support_code(self):
    return get_c_support_code_common() + get_c_support_code_mdlstm()

  def c_code(self, node, name, input_names, output_names, sub):
    X, W1, W2, W3, W4, V_h1, V_h2, V_h3, V_h4, V_v1, V_v2, V_v3, V_v4, \
        b1, b2, b3, b4, sizes, DY1, DY2, DY3, DY4, Y1, Y2, Y3, Y4, H1, H2, H3, H4 = input_names
    DX, DW1, DW2, DW3, DW4, DV_h1, DV_h2, DV_h3, DV_h4, \
        DV_v1, DV_v2, DV_v3, DV_v4, Db1, Db2, Db3, Db4 = output_names
    fail = sub['fail']
    inplace = "true" if self.inplace else "false"
    return """
    //std::cout << "MultiDirectionalTwoDLSTMOpGrad called" << std::endl;
    if(!%(inplace)s)
    {
      std::cout << "warning, inplace optimization failed, not working inplace" << std::endl;
    }

    if(%(DX)s || %(DW1)s || %(DW2)s || %(DW3)s || %(DW4)s ||
       %(DV_h1)s || %(DV_h2)s || %(DV_h3)s || %(DV_h4)s ||
       %(DV_v1)s || %(DV_v2)s || %(DV_v3)s || %(DV_v4)s ||
       %(Db1)s || %(Db2)s || %(Db3)s || %(Db4)s)
    {
      cout << "output storage already exists" << endl;
      //TODO check if we can reuse it
      Py_XDECREF(%(DX)s);
      Py_XDECREF(%(DW1)s);
      Py_XDECREF(%(DW2)s);
      Py_XDECREF(%(DW3)s);
      Py_XDECREF(%(DW4)s);
      Py_XDECREF(%(DV_h1)s);
      Py_XDECREF(%(DV_h2)s);
      Py_XDECREF(%(DV_h3)s);
      Py_XDECREF(%(DV_h4)s);
      Py_XDECREF(%(DV_v1)s);
      Py_XDECREF(%(DV_v2)s);
      Py_XDECREF(%(DV_v3)s);
      Py_XDECREF(%(DV_v4)s);
      Py_XDECREF(%(Db1)s);
      Py_XDECREF(%(Db2)s);
      Py_XDECREF(%(Db3)s);
      Py_XDECREF(%(Db4)s);
    }

    const int * X_dim = CudaNdarray_HOST_DIMS(%(X)s);
    const int * Y_dim = CudaNdarray_HOST_DIMS(%(Y1)s);
    const int height = X_dim[0];
    const int width = X_dim[1];
    const int n_minibatch = X_dim[2];
    const int n_diags = width + height - 1;
    const int max_diag_size = std::min(Y_dim[0], Y_dim[1]);

    CudaNdarray * delta1 = 0;
    CudaNdarray * delta2 = 0;
    CudaNdarray * delta3 = 0;
    CudaNdarray * delta4 = 0;
    if(%(inplace)s)
    {
      delta1 = %(H1)s;
      delta2 = %(H2)s;
      delta3 = %(H3)s;
      delta4 = %(H4)s;
    }
    else
    {
      delta1 = (CudaNdarray *) CudaNdarray_Copy(%(H1)s);
      delta2 = (CudaNdarray *) CudaNdarray_Copy(%(H2)s);
      delta3 = (CudaNdarray *) CudaNdarray_Copy(%(H3)s);
      delta4 = (CudaNdarray *) CudaNdarray_Copy(%(H4)s);
    }
    CudaNdarray * epsilon1 = (CudaNdarray *) CudaNdarray_Copy(%(DY1)s);
    CudaNdarray * epsilon2 = (CudaNdarray *) CudaNdarray_Copy(%(DY2)s);
    CudaNdarray * epsilon3 = (CudaNdarray *) CudaNdarray_Copy(%(DY3)s);
    CudaNdarray * epsilon4 = (CudaNdarray *) CudaNdarray_Copy(%(DY4)s);

    const int workmem1_dims[] = {2, Y_dim[0], Y_dim[1], Y_dim[2], Y_dim[3]};
    CudaNdarray * workmem1_1 = (CudaNdarray*) MyCudaNdarray_NewDims(5, workmem1_dims);
    assert(workmem1_1);
    CudaNdarray * workmem1_2 = (CudaNdarray*) MyCudaNdarray_NewDims(5, workmem1_dims);
    assert(workmem1_2);
    CudaNdarray * workmem1_3 = (CudaNdarray*) MyCudaNdarray_NewDims(5, workmem1_dims);
    assert(workmem1_3);
    CudaNdarray * workmem1_4 = (CudaNdarray*) MyCudaNdarray_NewDims(5, workmem1_dims);
    assert(workmem1_4);

    //we use floats to store float*'s, as CudaNdarray only supports floats. factor 10 for lstm bwd kernel
    int ptr_storage_dims[] = {4 * 10 * max_diag_size * sizeof(float*) / sizeof(float)};
    CudaNdarray * ptr_storage = (CudaNdarray*) MyCudaNdarray_NewDims(1, ptr_storage_dims);
    assert(ptr_storage);

    //valid: float tensor of 1s and 0s indicating the size of the image
    //4 dirs * max_diag_size * n_minibatch
    int valid_dims[] = {4 * max_diag_size * n_minibatch};
    CudaNdarray * valid_storage = (CudaNdarray*) MyCudaNdarray_NewDims(1, valid_dims);
    assert(valid_storage);

    for(int diag = n_diags-1; diag >= 0; --diag)
    {
      int diag_size = min(diag+1, min(abs(n_diags-diag), min(width, height)));
      int y_high = min(diag, height-1);
      int x_low = max(diag-height+1,0);
      vector<int> ys_h, xs_h, ys_v, xs_v, ys, xs;
      for(int idx = 0; idx < diag_size; ++idx)
      {
        int y = y_high - idx;
        int x = x_low + idx;
        bool rightBorder = (x == X_dim[1]-1);
        if(!rightBorder)
        {
          ys_h.push_back(y);
          xs_h.push_back(x);
        }
        bool botBorder = (y == X_dim[0]-1);
        if(!botBorder)
        {
          ys_v.push_back(y);
          xs_v.push_back(x);
        }
        ys.push_back(y);
        xs.push_back(x);
      }

      affine_y_x_batched_multidir(0, 1, delta1, delta2, delta3, delta4, %(V_h1)s, %(V_h2)s, %(V_h3)s, %(V_h4)s,
        epsilon1, epsilon2, epsilon3, epsilon4, ys_h, xs_h, ptr_storage, height, width, 0, false, true);
      affine_y_x_batched_multidir(1, 0, delta1, delta2, delta3, delta4, %(V_v1)s, %(V_v2)s, %(V_v3)s, %(V_v4)s,
        epsilon1, epsilon2, epsilon3, epsilon4, ys_v, xs_v, ptr_storage, height, width, 0, false, true);

      do_lstm_bwd_batched_multidir(delta1, delta2, delta3, delta4, epsilon1, epsilon2, epsilon3, epsilon4,
        %(Y1)s, %(Y2)s, %(Y3)s, %(Y4)s, workmem1_1, workmem1_2, workmem1_3, workmem1_4,
        X_dim[0], X_dim[1], ys, xs, ptr_storage, valid_storage, %(sizes)s);
    }
    Py_XDECREF(ptr_storage);
    Py_XDECREF(valid_storage);

    Py_XDECREF(workmem1_1);
    Py_XDECREF(workmem1_2);
    Py_XDECREF(workmem1_3);
    Py_XDECREF(workmem1_4);

    %(DX)s = CudaNdarray_uninitialized_like(%(X)s);
    assert(%(DX)s);
    %(DW1)s = CudaNdarray_uninitialized_like(%(W1)s);
    assert(%(DW1)s);
    %(DW2)s = CudaNdarray_uninitialized_like(%(W2)s);
    assert(%(DW2)s);
    %(DW3)s = CudaNdarray_uninitialized_like(%(W3)s);
    assert(%(DW3)s);
    %(DW4)s = CudaNdarray_uninitialized_like(%(W4)s);
    assert(%(DW4)s);

    //DW = X^T * delta
    affine_global(%(X)s, delta1, %(DW1)s, true, false, 0, 0, 0.0f);
    affine_global(%(X)s, delta2, %(DW2)s, true, false, 0, 0, 0.0f);
    affine_global(%(X)s, delta3, %(DW3)s, true, false, 0, 0, 0.0f);
    affine_global(%(X)s, delta4, %(DW4)s, true, false, 0, 0, 0.0f);
    //important! mind the order, first use X, then update DX, which might be aligned to X
    //DX = delta * W^T
    affine_global(delta1, %(W1)s, %(DX)s, false, true, 0, 0, 0.0f);
    affine_global(delta2, %(W2)s, %(DX)s, false, true, 0, 0, 1.0f);
    affine_global(delta3, %(W3)s, %(DX)s, false, true, 0, 0, 1.0f);
    affine_global(delta4, %(W4)s, %(DX)s, false, true, 0, 0, 1.0f);
    //Db = (1 ... 1) * delta
    %(Db1)s = sumOverAllButLastDimensions(delta1);
    %(Db2)s = sumOverAllButLastDimensions(delta2);
    %(Db3)s = sumOverAllButLastDimensions(delta3);
    %(Db4)s = sumOverAllButLastDimensions(delta4);

    //copy left/right part to workmem2 and set to 0 (could be done more efficient, but profiling shows, it's not worth it)
    const int * H_dim = CudaNdarray_HOST_DIMS(%(H1)s);
    const int workmem2_dims[] = {H_dim[0], H_dim[2], H_dim[3]};
    const int block_size = H_dim[2] * H_dim[3];
    CudaNdarray * workmem2_1 = (CudaNdarray*) MyCudaNdarray_NewDims(3, workmem2_dims);
    assert(workmem2_1);
    CudaNdarray * workmem2_2 = (CudaNdarray*) MyCudaNdarray_NewDims(3, workmem2_dims);
    assert(workmem2_2);
    CudaNdarray * workmem2_3 = (CudaNdarray*) MyCudaNdarray_NewDims(3, workmem2_dims);
    assert(workmem2_3);
    CudaNdarray * workmem2_4 = (CudaNdarray*) MyCudaNdarray_NewDims(3, workmem2_dims);
    assert(workmem2_4);
    for(int y = 0; y < Y_dim[0]; ++y)
    {
      float * workmem2_1_data_ptr = CudaNdarray_DEV_DATA(workmem2_1) + y * block_size;
      float * delta1_data_ptr = data_ptr(delta1, y, 0);
      HANDLE_ERROR(cudaMemcpy(workmem2_1_data_ptr, delta1_data_ptr, block_size * sizeof(float), cudaMemcpyDeviceToDevice));
      HANDLE_ERROR(cudaMemset(delta1_data_ptr, 0, sizeof(float) * H_dim[2] * H_dim[3]));

      float * workmem2_2_data_ptr = CudaNdarray_DEV_DATA(workmem2_2) + y * block_size;
      float * delta2_data_ptr = data_ptr(delta2, y, 0);
      HANDLE_ERROR(cudaMemcpy(workmem2_2_data_ptr, delta2_data_ptr, block_size * sizeof(float), cudaMemcpyDeviceToDevice));
      HANDLE_ERROR(cudaMemset(delta2_data_ptr, 0, sizeof(float) * H_dim[2] * H_dim[3]));

      float * workmem2_3_data_ptr = CudaNdarray_DEV_DATA(workmem2_3) + y * block_size;
      float * delta3_data_ptr = data_ptr(delta3, y, width - 1);
      HANDLE_ERROR(cudaMemcpy(workmem2_3_data_ptr, delta3_data_ptr, block_size * sizeof(float), cudaMemcpyDeviceToDevice));
      HANDLE_ERROR(cudaMemset(delta3_data_ptr, 0, sizeof(float) * H_dim[2] * H_dim[3]));

      float * workmem2_4_data_ptr = CudaNdarray_DEV_DATA(workmem2_4) + y * block_size;
      float * delta4_data_ptr = data_ptr(delta4, y, width - 1);
      HANDLE_ERROR(cudaMemcpy(workmem2_4_data_ptr, delta4_data_ptr, block_size * sizeof(float), cudaMemcpyDeviceToDevice));
      HANDLE_ERROR(cudaMemset(delta4_data_ptr, 0, sizeof(float) * H_dim[2] * H_dim[3]));
    }

    %(DV_h1)s = CudaNdarray_uninitialized_like(%(V_h1)s);
    assert(%(DV_h1)s);
    %(DV_h2)s = CudaNdarray_uninitialized_like(%(V_h2)s);
    assert(%(DV_h2)s);
    %(DV_h3)s = CudaNdarray_uninitialized_like(%(V_h3)s);
    assert(%(DV_h3)s);
    %(DV_h4)s = CudaNdarray_uninitialized_like(%(V_h4)s);
    assert(%(DV_h4)s);
    //DV_h = Y[0..end-1]^T * delta[1..end]
    affine_global(%(Y1)s, delta1, %(DV_h1)s, true, false, 0, 1, 0.0f);
    affine_global(%(Y2)s, delta2, %(DV_h2)s, true, false, 0, 1, 0.0f);
    affine_global(%(Y3)s, delta3, %(DV_h3)s, true, false, 1, 0, 0.0f);
    affine_global(%(Y4)s, delta4, %(DV_h4)s, true, false, 1, 0, 0.0f);

    //copy left/right part back
    for(int y = 0; y < Y_dim[0]; ++y)
    {
      float * workmem2_1_data_ptr = CudaNdarray_DEV_DATA(workmem2_1) + y * block_size;
      float * delta1_data_ptr = data_ptr(delta1, y, 0);
      HANDLE_ERROR(cudaMemcpy(delta1_data_ptr, workmem2_1_data_ptr, block_size * sizeof(float), cudaMemcpyDeviceToDevice));

      float * workmem2_2_data_ptr = CudaNdarray_DEV_DATA(workmem2_2) + y * block_size;
      float * delta2_data_ptr = data_ptr(delta2, y, 0);
      HANDLE_ERROR(cudaMemcpy(delta2_data_ptr, workmem2_2_data_ptr, block_size * sizeof(float), cudaMemcpyDeviceToDevice));

      float * workmem2_3_data_ptr = CudaNdarray_DEV_DATA(workmem2_3) + y * block_size;
      float * delta3_data_ptr = data_ptr(delta3, y, width - 1);
      HANDLE_ERROR(cudaMemcpy(delta3_data_ptr, workmem2_3_data_ptr, block_size * sizeof(float), cudaMemcpyDeviceToDevice));

      float * workmem2_4_data_ptr = CudaNdarray_DEV_DATA(workmem2_4) + y * block_size;
      float * delta4_data_ptr = data_ptr(delta4, y, width - 1);
      HANDLE_ERROR(cudaMemcpy(delta4_data_ptr, workmem2_4_data_ptr, block_size * sizeof(float), cudaMemcpyDeviceToDevice));
    }

    Py_XDECREF(workmem2_1);
    Py_XDECREF(workmem2_2);
    Py_XDECREF(workmem2_3);
    Py_XDECREF(workmem2_4);

    %(DV_v1)s = CudaNdarray_uninitialized_like(%(V_v1)s);
    assert(%(DV_v1)s);
    %(DV_v2)s = CudaNdarray_uninitialized_like(%(V_v2)s);
    assert(%(DV_v2)s);
    %(DV_v3)s = CudaNdarray_uninitialized_like(%(V_v3)s);
    assert(%(DV_v3)s);
    %(DV_v4)s = CudaNdarray_uninitialized_like(%(V_v4)s);
    assert(%(DV_v4)s);

    //DV_v = Y[0..end-1]^T * delta[1..end]
    affine_global(%(Y1)s, delta1, %(DV_v1)s, true, false, 0, Y_dim[1], 0.0f);
    affine_global(%(Y2)s, delta2, %(DV_v2)s, true, false, Y_dim[1], 0, 0.0f);
    affine_global(%(Y3)s, delta3, %(DV_v3)s, true, false, 0, Y_dim[1], 0.0f);
    affine_global(%(Y4)s, delta4, %(DV_v4)s, true, false, Y_dim[1], 0, 0.0f);

    //for debugging
    /*cout << "=====delta1====" << endl;
    CudaNdarray_print_part(delta1);
    cout << "=====delta2====" << endl;
    CudaNdarray_print_part(delta2);
    cout << "=====delta3====" << endl;
    CudaNdarray_print_part(delta3);
    cout << "=====delta4====" << endl;
    CudaNdarray_print_part(delta4);
    cout << "===============" << endl;*/

    if(!%(inplace)s)
    {
      Py_XDECREF(delta1);
      Py_XDECREF(delta2);
      Py_XDECREF(delta3);
      Py_XDECREF(delta4);
    }
    Py_XDECREF(epsilon1);
    Py_XDECREF(epsilon2);
    Py_XDECREF(epsilon3);
    Py_XDECREF(epsilon4);

    """ % locals()

  #!!! change this when changing the code!
  def c_code_cache_version(self):
    return 2, 10

MultiDirectionalTwoDLSTMOpGradNoInplaceInstance = MultiDirectionalTwoDLSTMOpGrad(inplace=False)
MultiDirectionalTwoDLSTMOpGradInplaceInstance = MultiDirectionalTwoDLSTMOpGrad(inplace=True)

MultiDirectionalTwoDLSTMOpInplaceOpt = OpSub(MultiDirectionalTwoDLSTMOpGradNoInplaceInstance,
                                             MultiDirectionalTwoDLSTMOpGradInplaceInstance)

#hack to avoid being called twice
if not hasattr(optdb, 'MultiDirectionalTwoDLSTMOpInplaceOpt_registered'):
  optdb.register('MultiDirectionalTwoDLSTMOpInplaceOpt',
                 theano.gof.TopoOptimizer(MultiDirectionalTwoDLSTMOpInplaceOpt, failure_callback=gof.TopoOptimizer.warn_inplace),
                 50.0, 'fast_run', 'inplace', 'gpuarray')
  optdb.MultiDirectionalTwoDLSTMOpInplaceOpt_registered = True


class MultiDirectionalTwoDLSTMOp(theano.sandbox.cuda.GpuOp):
  __props__ = ()

  def __init__(self):
    super(MultiDirectionalTwoDLSTMOp, self).__init__()

  def make_node(self, X, W1, W2, W3, W4, V_h1, V_h2, V_h3, V_h4, V_v1, V_v2, V_v3, V_v4, b1, b2, b3, b4, sizes):
    var_names = ["X", "W1", "W2", "W3", "W4", "V_h1", "V_h2", "V_h3", "V_h4",
                 "V_v1", "V_v2", "V_v3", "V_v4", "b1", "b2", "b3", "b4"]
    lcl = locals()
    for var_name in var_names:
      lcl[var_name] = gpu_contiguous(as_cuda_ndarray_variable(lcl[var_name]))
      assert lcl[var_name].dtype == "float32"
    #note: sizes lives on the CPU!
    sizes = T.as_tensor_variable(sizes)
    assert sizes.dtype == "float32"

    assert lcl["X"].ndim == 4
    assert lcl["W1"].ndim == 2
    assert lcl["W2"].ndim == 2
    assert lcl["W3"].ndim == 2
    assert lcl["W4"].ndim == 2
    assert lcl["V_h1"].ndim == 2
    assert lcl["V_h2"].ndim == 2
    assert lcl["V_h3"].ndim == 2
    assert lcl["V_h4"].ndim == 2
    assert lcl["V_v1"].ndim == 2
    assert lcl["V_v2"].ndim == 2
    assert lcl["V_v3"].ndim == 2
    assert lcl["V_v4"].ndim == 2
    assert lcl["b1"].ndim == 1
    assert lcl["b2"].ndim == 1
    assert lcl["b3"].ndim == 1
    assert lcl["b4"].ndim == 1
    assert sizes.ndim == 2

    all_vars = [lcl[var_name] for var_name in var_names] + [sizes]

    #results: outputs Y1, Y2, Y3, Y4, (gates and cell states) H1, H2, H3, H4
    return theano.Apply(self, all_vars, [lcl["X"].type() for _ in range(8)])

  def c_support_code(self):
    return get_c_support_code_common() + get_c_support_code_mdlstm()

  def c_code(self, node, name, input_names, output_names, sub):
    X, W1, W2, W3, W4, V_h1, V_h2, V_h3, V_h4, V_v1, V_v2, V_v3, V_v4, b1, b2, b3, b4, sizes = input_names
    Y1, Y2, Y3, Y4, H1, H2, H3, H4 = output_names
    fail = sub['fail']
    return """
    //std::cout << "MultiDirectionalTwoDLSTMOp called" << std::endl;
    if(%(Y1)s || %(Y2)s || %(Y3)s || %(Y4)s || %(H1)s || %(H2)s || %(H3)s || %(H4)s)
    {
      //cout << "Ys or Hs already exist" << endl;
      //TODO check if we can reuse it
      Py_XDECREF(%(Y1)s);
      Py_XDECREF(%(Y2)s);
      Py_XDECREF(%(Y3)s);
      Py_XDECREF(%(Y4)s);
      Py_XDECREF(%(H1)s);
      Py_XDECREF(%(H2)s);
      Py_XDECREF(%(H3)s);
      Py_XDECREF(%(H4)s);
    }

    const int * X_dim = CudaNdarray_HOST_DIMS(%(X)s);
    const int * W_dim = CudaNdarray_HOST_DIMS(%(W1)s);
    const int * V_dim = CudaNdarray_HOST_DIMS(%(V_h1)s);
    assert(W_dim[1] %% 5 == 0 && "W has wrong shape");
    assert(5 * V_dim[0] == V_dim[1] && "V has wrong shape");
    assert(W_dim[1] == V_dim[1]);
    assert(W_dim[0] == X_dim[3]);
    const int Y_dim[] = {X_dim[0], X_dim[1], X_dim[2], W_dim[1] / 5};
    const int H_dim[] = {X_dim[0], X_dim[1], X_dim[2], W_dim[1]};
    const int height = X_dim[0];
    const int width = X_dim[1];
    const int n_minibatch = X_dim[2];
    const int max_diag_size = std::min(height, width);
    const int n_diags = width + height - 1;

    //init Ys
    %(Y1)s = (CudaNdarray*) MyCudaNdarray_NewDims(4, Y_dim);
    assert(%(Y1)s);
    %(Y2)s = (CudaNdarray*) MyCudaNdarray_NewDims(4, Y_dim);
    assert(%(Y2)s);
    %(Y3)s = (CudaNdarray*) MyCudaNdarray_NewDims(4, Y_dim);
    assert(%(Y3)s);
    %(Y4)s = (CudaNdarray*) MyCudaNdarray_NewDims(4, Y_dim);
    assert(%(Y4)s);

    //init Hs
    %(H1)s = (CudaNdarray*) MyCudaNdarray_NewDims(4, H_dim);
    assert(%(H1)s);
    %(H2)s = (CudaNdarray*) MyCudaNdarray_NewDims(4, H_dim);
    assert(%(H2)s);
    %(H3)s = (CudaNdarray*) MyCudaNdarray_NewDims(4, H_dim);
    assert(%(H3)s);
    %(H4)s = (CudaNdarray*) MyCudaNdarray_NewDims(4, H_dim);
    assert(%(H4)s);

    //init Hs with bs
    fillmat(%(b1)s, %(H1)s);
    fillmat(%(b2)s, %(H2)s);
    fillmat(%(b3)s, %(H3)s);
    fillmat(%(b4)s, %(H4)s);

    //H+=XW
    affine_global(%(X)s, %(W1)s, %(H1)s);
    affine_global(%(X)s, %(W2)s, %(H2)s);
    affine_global(%(X)s, %(W3)s, %(H3)s);
    affine_global(%(X)s, %(W4)s, %(H4)s);

    //we use floats to store float*'s, as CudaNdarray only supports floats. factor 5 for lstm kernel,
    //additional factor 4 for 4 directions
    int ptr_storage_dims[] = {4 * 5 * max_diag_size * sizeof(float*) / sizeof(float)};
    CudaNdarray * ptr_storage = (CudaNdarray*) MyCudaNdarray_NewDims(1, ptr_storage_dims);
    assert(ptr_storage);

    //valid: float tensor of 1s and 0s indicating the size of the image
    //4 dirs * max_diag_size * n_minibatch
    int valid_dims[] = {4 * max_diag_size * n_minibatch};
    CudaNdarray * valid_storage = (CudaNdarray*) MyCudaNdarray_NewDims(1, valid_dims);
    assert(valid_storage);

    for(int diag = 0; diag < n_diags; ++diag)
    {
      int diag_size = min(diag+1, min(abs(n_diags-diag), min(width, height)));
      int y_high = min(diag, height-1);
      int x_low = max(diag-height+1,0);
      vector<int> ys_h, xs_h, ys_v, xs_v, ys, xs;
      for(int idx = 0; idx < diag_size; ++idx)
      {
        int y = y_high - idx;
        int x = x_low + idx;
        if(x > 0)
        {
          ys_h.push_back(y);
          xs_h.push_back(x);
        }
        if(y > 0)
        {
          ys_v.push_back(y);
          xs_v.push_back(x);
        }
        ys.push_back(y);
        xs.push_back(x);
      }

      affine_y_x_batched_multidir(0, -1,
        %(Y1)s, %(Y2)s, %(Y3)s, %(Y4)s,
        %(V_h1)s, %(V_h2)s, %(V_h3)s, %(V_h4)s,
        %(H1)s, %(H2)s, %(H3)s, %(H4)s,
        ys_h, xs_h, ptr_storage, height, width);
      affine_y_x_batched_multidir(-1, 0,
        %(Y1)s, %(Y2)s, %(Y3)s, %(Y4)s,
        %(V_v1)s, %(V_v2)s, %(V_v3)s, %(V_v4)s,
        %(H1)s, %(H2)s, %(H3)s, %(H4)s,
        ys_v, xs_v, ptr_storage, height, width);

      do_lstm_batched_multidir(%(H1)s, %(H2)s, %(H3)s, %(H4)s,
                                 %(Y1)s, %(Y2)s, %(Y3)s, %(Y4)s,
                                 ys, xs, ptr_storage, valid_storage, %(sizes)s);
    }
    Py_XDECREF(ptr_storage);
    Py_XDECREF(valid_storage);

    """ % locals()

  def grad(self, inputs, output_grads):
    raw_inputs = [raw_variable(v) for v in inputs]
    fwd_results = self(*raw_inputs)
    args = inputs + output_grads[:4] + fwd_results
    grads = MultiDirectionalTwoDLSTMOpGradNoInplaceInstance(*args)
    Dsizes = theano.gradient.grad_undefined(self, len(inputs) - 1, inputs[-1], 'cannot diff w.r.t. sizes')
    return grads + [Dsizes]

  # noinspection PyMethodMayBeStatic
  def infer_shape(self, node, input_shapes):
    Xs, W1s = input_shapes[:2]
    Y_shape = (Xs[0], Xs[1], Xs[2], W1s[1] // 5)
    H_shape = (Xs[0], Xs[1], Xs[2], W1s[1])
    return [Y_shape, Y_shape, Y_shape, Y_shape, H_shape, H_shape, H_shape, H_shape]

  #!!! change this when changing the code!
  def c_code_cache_version(self):
    return 2, 10

MultiDirectionalTwoDLSTMOpInstance = MultiDirectionalTwoDLSTMOp()




from __future__ import print_function

import _setup_test_env  # noqa
from subprocess import Popen, PIPE, STDOUT, CalledProcessError
import re
import os
import sys
from glob import glob
import unittest
from nose.tools import assert_less, assert_in
from returnn.util import better_exchook
from returnn.util.basic import which_pip


py = sys.executable
print("Python:", py)

os.chdir((os.path.dirname(__file__) or ".") + "/..")
assert os.path.exists("rnn.py")


def build_env(env_update=None):
  """
  :param dict[str,str]|None env_update:
  :return: env dict for Popen
  :rtype: dict[str,str]
  """
  env_update_ = os.environ.copy()
  if env_update:
    env_update_.update(env_update)
  return env_update_


def run(*args, env_update=None):
  args = list(args)
  print("run:", args)
  # RETURNN by default outputs on stderr, so just merge both together
  p = Popen(args, stdout=PIPE, stderr=STDOUT, env=build_env(env_update=env_update))
  out, _ = p.communicate()
  if p.returncode != 0:
    print("Return code is %i" % p.returncode)
    print("std out/err:\n---\n%s\n---\n" % out.decode("utf8"))
    raise CalledProcessError(cmd=args, returncode=p.returncode, output=out)
  return out.decode("utf8")


def run_and_parse_last_fer(*args, **kwargs):
  out = run(*args, **kwargs)
  parsed_fer = None
  for line in out.splitlines():
    # example: epoch 5 score: 0.0231807245472 elapsed: 0:00:04 dev: score 0.0137521058997 error 0.00268961807423
    m = re.match("epoch [0-9]+ score: .* dev: .* error ([0-9.]+)\\s?", line)
    if not m:
      # example: dev: score 0.03350000149202181 error 0.009919877954075871
      m = re.match("dev: score .* error ([0-9.]+)\\s?", line)
    if not m:
      continue
    parsed_fer = float(m.group(1))
  err_msg = "ERROR: No epoch dev errors found in output"
  assert parsed_fer is not None, "%s.\nOutput:\n\n%s\n\n%s." % (err_msg, out, err_msg)
  return parsed_fer


def run_config_get_fer(config_filename, env_update=None):
  cleanup_tmp_models(config_filename)
  fer = run_and_parse_last_fer(
    py, "rnn.py", config_filename, "++log_verbosity", "5", env_update=env_update)
  print("FER:", fer)
  cleanup_tmp_models(config_filename)
  return fer


def cleanup_tmp_models(config_filename):
  assert os.path.exists(config_filename)
  from returnn.config import Config
  config = Config()
  config.load_file(config_filename)
  model_filename = config.value('model', '')
  assert model_filename
  # Remove existing models
  assert model_filename.startswith("/tmp/")
  for f in glob(model_filename + ".*"):
    os.remove(f)


def test_demo_tf_task12ax():
  fer = run_config_get_fer("demos/demo-tf-native-lstm.12ax.config")
  assert_less(fer, 0.01)


def test_demo_tf_task12ax_no_test_env():
  fer = run_config_get_fer("demos/demo-tf-native-lstm2.12ax.config", env_update={"RETURNN_TEST": ""})
  assert_less(fer, 0.01)


def test_demo_iter_dataset_task12ax():
  cleanup_tmp_models("demos/demo-theano-task12ax.config")
  # While the config is for Theano, it should not matter for demo-iter-dataset
  out = run(py, "demos/demo-iter-dataset.py", "demos/demo-theano-task12ax.config")
  assert_in("Epoch 5.", out.splitlines())


def test_demo_returnn_as_framework():
  print("Prepare.")
  import subprocess
  import shutil
  from glob import glob
  from returnn.util.basic import get_login_username
  # echo via subprocess, because this stdout as well as the other will always be visible.
  subprocess.check_call(["echo", "travis_fold:start:test_demo_returnn_as_framework"])
  assert os.path.exists("setup.py")
  if glob("dist/*.tar.gz"):
    # we want it unique below
    for fn in glob("dist/*.tar.gz"):
      os.remove(fn)
  if os.path.exists("MANIFEST"):
    os.remove("MANIFEST")  # auto-generated. will be recreated
  if os.path.exists("docs/crnn"):
    os.remove("docs/crnn")  # this is auto-generated, and confuses setup.py sdist
  if os.path.exists("docs/returnn"):
    os.remove("docs/returnn")  # this is auto-generated, and confuses setup.py sdist
  tmp_model_dir = "/tmp/%s/returnn-demo-as-framework" % get_login_username()
  if os.path.exists(tmp_model_dir):
    shutil.rmtree(tmp_model_dir, ignore_errors=True)
  print("setup.py sdist, to create package.")
  subprocess.check_call([py, "setup.py", "sdist"])
  dist_fns = glob("dist/*.tar.gz")
  assert len(dist_fns) == 1
  dist_fn = os.path.abspath(dist_fns[0])
  pip_path = which_pip()
  print("Pip install Returnn.")
  in_virtual_env = hasattr(sys, 'real_prefix')  # https://stackoverflow.com/questions/1871549/
  cmd = [py, pip_path, "install"]
  if not in_virtual_env:
    cmd += ["--user"]
  cmd += ["-v", dist_fn]
  print("$ %s" % " ".join(cmd))
  subprocess.check_call(cmd, cwd="/")
  print("Running demo now.")
  subprocess.check_call([py, "demo-returnn-as-framework.py"], cwd="demos")
  print("Success.")
  subprocess.check_call(["echo", "travis_fold:end:test_demo_returnn_as_framework"])


def test_demo_sprint_interface():
  import subprocess
  # echo via subprocess, because this stdout as well as the other will always be visible.
  subprocess.check_call(["echo", "travis_fold:start:test_demo_sprint_interface"])
  subprocess.check_call([py, os.path.abspath("demos/demo-sprint-interface.py")], cwd="/")
  subprocess.check_call(["echo", "travis_fold:end:test_demo_sprint_interface"])


def test_returnn_as_framework_TaskSystem():
  import subprocess
  # echo via subprocess, because this stdout as well as the other will always be visible.
  subprocess.check_call(["echo", "travis_fold:start:test_returnn_as_framework_TaskSystem"])
  subprocess.check_call([py, os.path.abspath("tests/returnn-as-framework.py"), "test_TaskSystem_Pickler()"], cwd="/")
  subprocess.check_call(["echo", "travis_fold:end:test_returnn_as_framework_TaskSystem"])


def test_returnn_as_framework_old_style_crnn_TFUtil():
  import subprocess
  # echo via subprocess, because this stdout as well as the other will always be visible.
  subprocess.check_call(["echo", "travis_fold:start:test_returnn_as_framework_old_style_crnn_TFUtil"])
  subprocess.check_call([
    py, os.path.abspath("tests/returnn-as-framework.py"),
    "--old-style", "--returnn-package-name", "crnn",
    "test_old_style_import_crnn_TFUtil()"], cwd="/")
  subprocess.check_call(["echo", "travis_fold:end:test_returnn_as_framework_old_style_crnn_TFUtil"])


def test_returnn_as_framework_old_style_TFUtil():
  import subprocess
  # echo via subprocess, because this stdout as well as the other will always be visible.
  subprocess.check_call(["echo", "travis_fold:start:test_returnn_as_framework_old_style_TFUtil"])
  subprocess.check_call([
    py, os.path.abspath("tests/returnn-as-framework.py"), "--old-style", "test_old_style_import_TFUtil()"], cwd="/")
  subprocess.check_call(["echo", "travis_fold:end:test_returnn_as_framework_old_style_TFUtil"])


if __name__ == '__main__':
  better_exchook.install()
  import sys
  fns = sys.argv[1:]
  if not fns:
    fns = [arg for arg in globals() if arg.startswith("test_")]
  for arg in fns:
    f = globals()[arg]
    print("-" * 20)
    print("Run:", f)
    print("-" * 20)
    try:
      f()
    except unittest.SkipTest as exc:
      print("(SkipTest: %s)" % exc)
    print("-" * 20)
    print("Ok.")
    print("-" * 20)



"""
The Server module manages connections via HTTP. It is especially
tuned to work well with multiple engine instances at once. Initiated
via rnn.py with a Server config. Here's a basic config:

#!returnn/rnn.py
task = "server"
port = 10687
max_engines = 2
network = {"out" : { "class" : "softmax", "loss" : "ce", "target":"classes" }}
"""

from __future__ import print_function

from array import array
import concurrent.futures
import datetime
import hashlib
import json
import os
import re
import struct
import sys
import time
import urllib
import urllib.request

import numpy as np
from tornado import locks
from tornado.concurrent import Future
from tornado.ioloop import IOLoop
from tornado.queues import Queue, QueueEmpty
import tornado.web

from returnn.log import log
from returnn.datasets.generating import StaticDataset
from returnn.theano.device import Device
from returnn.util.basic import TheanoFlags
from returnn.config import get_devices_init_args
from returnn.theano.engine_task import ForwardTaskThread
import returnn.theano.engine
import returnn.config

# TODO: Look into making these non-global.
_max_amount_engines = 4


class ClassificationRequest:
  def __init__(self, data):
    self.data = data
    self.future = Future()


class Model:
  def __init__(self, config_file):
    self.lock = locks.Lock()
    self.classification_queue = Queue()

    print('loading config %s' % config_file, file=log.v5)
    # Load and setup config
    try:
      self.config = returnn.config.Config()
      self.config.load_file(config_file)
      self.pause_after_first_seq = self.config.float('pause_after_first_seq', 0.2)
      self.batch_size = self.config.int('batch_size', 5000)
      self.max_seqs = self.config.int('max_seqs', -1)
    except Exception:
      print('Error: loading config %s failed' % config_file, file=log.v1)
      raise

    try:
      self.devices = self._init_devices()
    except Exception:
      print('Error: Loading devices for config %s failed' % config_file, file=log.v1)
      raise

    print('Starting engine for config %s' % config_file, file=log.v5)
    self.engine = returnn.theano.engine.Engine(self.devices)
    try:
      self.engine.init_network_from_config(config=self.config)
    except Exception:
      print('Error: Loading network for config %s failed' % config_file, file=log.v1)
      raise

    IOLoop.current().spawn_callback(self.classify_in_background)

    self.last_used = datetime.datetime.now()

  def _init_devices(self):
    """
    Initiates the required devices for a config. Same as the funtion initDevices in
    rnn.py.
    :param config:
    :return: A list with the devices used.
    """
    oldDeviceConfig = ",".join(self.config.list('device', ['default']))
    if "device" in TheanoFlags:
      # This is important because Theano likely already has initialized that device.
      config.set("device", TheanoFlags["device"])
      print("Devices: Use %s via THEANO_FLAGS instead of %s." % (TheanoFlags["device"], oldDeviceConfig), file=log.v4)
    devArgs = get_devices_init_args(self.config)
    assert len(devArgs) > 0
    devices = [Device(**kwargs) for kwargs in devArgs]
    for device in devices:
      while not device.initialized:
        time.sleep(0.25)
    if devices[0].blocking:
      print("Devices: Used in blocking / single proc mode.", file=log.v4)
    else:
      print("Devices: Used in multiprocessing mode.", file=log.v4)
    return devices

  @tornado.gen.coroutine
  def classify_in_background(self):
    while True:
      requests = []
      # fetch first request
      r = yield self.classification_queue.get()
      requests.append(r)
      # grab all other waiting requests
      try:
        while True:
          requests.append(self.classification_queue.get_nowait())
      except QueueEmpty:
        pass

      output_dim = {}
      # Do dataset creation and classification.
      dataset = StaticDataset(data=[r.data for r in requests], output_dim=output_dim)
      dataset.init_seq_order()
      batches = dataset.generate_batches(recurrent_net=self.engine.network.recurrent,
                                         batch_size=self.batch_size, max_seqs=self.max_seqs)

      with (yield self.lock.acquire()):
        ctt = ForwardTaskThread(self.engine.network, self.devices, dataset, batches)
        yield ctt.join()

      try:
        for i in range(dataset.num_seqs):
          requests[i].future.set_result(ctt.result[i])
          self.classification_queue.task_done()
      except Exception as e:
        print('exception', e)
        raise

  @tornado.gen.coroutine
  def classify(self, data):
    self.last_used = datetime.datetime.now()
    request = ClassificationRequest(data)

    yield self.classification_queue.put(request)
    yield request.future

    # return request.future.result()


class Server:
  def __init__(self, global_config):
    """
    Initializes the server with an empty config.
    :param global_config: Basic config of server. Requires a network paramater.
    """

    # Create temporary directories.
    self.base_dir   = os.path.abspath(returnn.config.get_global_config().value('__file__', ''))
    self.config_dir = os.path.join(self.base_dir, "configs")
    self.data_dir   = os.path.join(self.base_dir, "data")

    for d in [self.config_dir, self.data_dir]:
      try:
        os.makedirs(d)
      except FileExistsError as e:
        pass

    self.models = {}

    self.application = tornado.web.Application([
      (r"/classify", ClassifyHandler, {'models': self.models}),
      (r"/loadconfig", ConfigHandler, {'config_dir': self.config_dir, 'models': self.models})
    ], debug=True, autoreload=False)

    self.port = int(global_config.value('port', '3033'))
    global _max_amount_engines
    _max_amount_engines = int(global_config.value('max_engines', '5'))

  def run(self):
    print("Starting server on port: %d" % self.port, file=log.v3)
    self.application.listen(self.port)
    IOLoop.instance().start()


class ClassifyHandler(tornado.web.RequestHandler):
  def initialize(self, models, **kwargs):
    self.models = models

  @tornado.web.asynchronous
  @tornado.gen.coroutine
  def post(self, *args, **kwargs):
    # TODO: Write formal documentation
    """
    Method for handling classification via HTTP Post request. The following must
    be defined in the URL paramaters: engine_hash (engine hash which points to which
    engine to use), and the data itself in the body. If using binary data, the following
    URL paramaters must also be supplied: data_format='binary', data_shape=(<dim1,dim2>).
    If using a specific data type, you can supply it as the url parameter data_type.
    :param args:
    :param kwargs:
    :return: Either JSON with error or JSON list of generated outputs.
    """
    url_params = self.request.arguments
    output_dim = {}
    ret = {}
    data = {}
    data_format = 'json'
    data_type = 'float32'
    engine_hash = ''
    data_shape = ''
    # First get meta data from URL parameters
    if 'engine_hash' in url_params:
      engine_hash = url_params['engine_hash'][0].decode('utf8')
    if 'data_format' in url_params:
      data_format = url_params['data_format'][0].decode('utf8')
    if 'data_type' in url_params:
      # Possible options: https://docs.scipy.org/doc/numpy-1.10.1/user/basics.types.html
      data_type = url_params['data_type'][0].decode('utf8')
    if 'data_shape' in url_params:
      data_shape = url_params['data_shape'][0].decode('utf8')  # either '' or 'dim1,dim2'

    print('Received request, engine hash: %s, data format: %s, data type %s data shape: %s' %
          (engine_hash, data_format, data_type, data_shape), file=log.v5)
    # Load in engine and hash
    try:
      model = self.models[engine_hash]
    except KeyError:
      print('Error: Received request with unknown engine_hash %s' % engine_hash, file=log.v1)
      self.set_status(499, 'Unknown engine_hash')
      return
    network = model.engine.network

    # Pre-process the data
    if data_format == 'json':
      data = json.loads(self.request.body)
      for k in data:
        try:
          data[k] = np.asarray(data[k], dtype=data_type)
          if k != 'data':
            output_dim[k] = network.n_out[k]  # = [network.n_in,2] if k == 'data' else network.n_out[k]
        except Exception:
          if k != 'data' and k not in network.n_out:
            self.set_status(499, 'unknown target: %s' % k)
          else:
            self.set_status(499, 'unable to convert %s to an array from value %s' % (k, str(data[k])))
          return
    elif data_format == 'binary':
      float_array = array(self._get_type_code(data_type))
      try:
        float_array.frombytes(self.request.body)
      except Exception as e:
        print('Binary data error: %s' % str(e.message), file=log.v4)
        self.set_status(499, 'Error during binary data conversion: ' + e.message)
        return
      shape = tuple(map(int, data_shape.split(',')))
      data['data'] = np.asarray(float_array.tolist(), dtype=data_type).reshape(shape)

    result = yield model.classify(data)
    print('writing results for request with shape %s' % data_shape)
    if data_format == 'json':
      self.write({'result': result.tolist()})
    elif data_format == 'binary':
      size_info = struct.pack('<LL', *result.shape)
      self.write(size_info)
      self.write(result.tobytes())


  def _get_type_code(self, format_to_convert):
    """
    Converts from Numpy format to python internal array format.
    :param format_to_convert: String of format used, e.g float32.
    :return: Returns a string of the type used for python arrays.
    """
    return {
      'float32': 'f',
      'float64': 'd',
      'int8': 'h',
      'int16': 'i',
      'int32': 'l',
    }[format_to_convert]


class ConfigHandler(tornado.web.RequestHandler):
  load_config_lock = locks.Lock()

  def initialize(self, config_dir, models, **kwargs):
    self.config_dir = config_dir
    self.models     = models

  @tornado.web.asynchronous
  @tornado.gen.coroutine
  def post(self, *args, **kwargs):
    """
    Handles the creation of a new engine based on a slightly modified config, supplied
    via HTTP Post. The request requires 1 URL parameter: new_config_url, which points
    to a URL (can be local) from where to download the config. This call is blocking.

    WARNING All configs must have the following added:
    extract_output_layer_name = "<OUTPUT LAYER ID>"
    :param args:
    :param kwargs:
    :return: ASCII encoded hash of the new engine based on the provided config.
    """
    data = json.loads(self.request.body)
    config_url = data["new_config_url"]
    print('Received request to load config %s' % config_url, file=log.v3)

    hash_engine = hashlib.new('ripemd160')
    hash_engine.update(config_url.encode('utf8'))
    hash_val = hash_engine.hexdigest()

    with (yield self.load_config_lock.acquire()):
      if hash_val in self.models:
        print('Using existing model with hash %s' % hash_val, file=log.v3)
        self.write(hash_val)
        return

      # Download new config file and save to temp folder.
      print('loading config %s' % config_url, file=log.v5)
      config_file = os.path.join(self.config_dir, hash_val + ".config")
      try:
        urllib.request.urlretrieve(config_url, config_file)
      except (urllib.request.URLError, urllib.request.HTTPError) as e:
        self.write('Error: Loading in config file from URL %s' % e)
        return

      try:
        self.models[hash_val] = Model(config_file)
      except Exception as e:
        print('Error: loading model failed %s' % str(e), file=log.v1)
        self.set_status(499, str(e))
      else:
        self.write(hash_val)


# There used to be a training handler, but it was not finished and the server needed refactoring, so it was deleted to save some time. If you wish to revive it look in the VCS history for its code
